# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import time
from typing import List

from nlp_architect.common.cdc.mention_data import MentionData
from nlp_architect.utils.io import load_json_file

logger = logging.getLogger(__name__)


class Topic(object):
    def __init__(self, topic_id):
        self.topic_id = topic_id
        self.mentions = []


class Topics(object):
    def __init__(self):
        self.topics_list = []
        self.keep_order = False

    def create_from_file(self, mentions_file_path: str, keep_order: bool = False) -> None:
        """

        Args:
            keep_order: whether to keep original mentions order or not (default = False)
            mentions_file_path: this topic mentions json file
        """
        self.keep_order = keep_order
        self.topics_list = self.load_mentions_from_file(mentions_file_path)

    def load_mentions_from_file(self, mentions_file_path: str) -> List[Topic]:
        start_data_load = time.time()
        logger.info('Loading mentions from-%s', mentions_file_path)
        mentions = load_json_file(mentions_file_path)
        topics = self.order_mentions_by_topics(mentions)
        end_data_load = time.time()
        took_load = end_data_load - start_data_load
        logger.info('Mentions file-%s, took:%.4f sec to load', mentions_file_path, took_load)
        return topics

    def order_mentions_by_topics(self, mentions: str) -> List[Topic]:
        """
        Order mentions to documents topics
        Args:
            mentions: json mentions file

        Returns:
            List[Topic] of the mentions separated by their documents topics
        """
        running_index = 0
        topics = []
        current_topic_ref = None
        for mention_line in mentions:
            mention = MentionData.read_json_mention_data_line(mention_line)

            if self.keep_order:
                if mention.mention_index == -1:
                    mention.mention_index = running_index
                    running_index += 1

            topic_id = mention.topic_id

            if not current_topic_ref or len(topics) > 0 and topic_id != topics[-1].topic_id:
                current_topic_ref = Topic(topic_id)
                topics.append(current_topic_ref)

            current_topic_ref.mentions.append(mention)

        return topics



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import absolute_import, division, print_function, unicode_literals

import tensorflow as tf

from nlp_architect.contrib.tensorflow.python.keras.layers.crf import CRF
from nlp_architect.contrib.tensorflow.python.keras.utils.layer_utils import load_model, save_model


class IntentExtractionModel(object):
    """
    Intent Extraction model base class (using tf.keras)
    """

    def __init__(self):
        self.model = None

    def fit(self, x, y, epochs=1, batch_size=1, callbacks=None, validation=None):
        """
        Train a model given input samples and target labels.

        Args:
            x: input samples
            y: input sample labels
            epochs (:obj:`int`, optional): number of epochs to train
            batch_size (:obj:`int`, optional): batch size
            callbacks(:obj:`Callback`, optional): Keras compatible callbacks
            validation(:obj:`list` of :obj:`numpy.ndarray`, optional): optional validation data
                to be evaluated when training
        """
        assert self.model, 'Model was not initialized'
        self.model.fit(x, y, epochs=epochs, batch_size=batch_size, shuffle=True,
                       validation_data=validation,
                       callbacks=callbacks)

    def predict(self, x, batch_size=1):
        """
        Get the prediction of the model on given input

        Args:
            x: samples to run through the model
            batch_size (:obj:`int`, optional): batch size:

        Returns:
            numpy.ndarray: predicted values by the model
        """
        assert self.model, 'Model was not initialized'
        return self.model.predict(x, batch_size=batch_size)

    def save(self, path, exclude=None):
        """
        Save model to path

        Args:
            path (str): path to save model
            exclude (list, optional): a list of object fields to exclude when saving
        """
        assert self.model, 'Model was not initialized'
        topology = {k: v for k, v in self.__dict__.items()}
        topology.pop('model')
        if exclude and isinstance(exclude, list):
            for x in exclude:
                topology.pop(x)
        save_model(self.model, topology=topology, filepath=path)

    def load(self, path):
        """
        Load a trained model

        Args:
            path (str): path to model file
        """
        load_model(path, self)

    @property
    def input_shape(self):
        """:obj:`tuple`:Get input shape"""
        return self.model.layers[0].input_shape

    @staticmethod
    def _create_input_embed(sentence_len, is_extern_emb, token_emb_size, vocab_size):
        if is_extern_emb:
            in_layer = e_layer = tf.keras.layers.Input(shape=(sentence_len, token_emb_size,),
                                                       dtype='float32', name='tokens_input')
        else:
            in_layer = tf.keras.layers.Input(shape=(sentence_len,),
                                             dtype='int32', name='tokens_input')
            e_layer = tf.keras.layers.Embedding(vocab_size, token_emb_size,
                                                input_length=sentence_len,
                                                name='embedding_layer')(in_layer)
        return in_layer, e_layer

    def load_embedding_weights(self, weights):
        """
        Load word embedding weights into the model embedding layer

        Args:
            weights (numpy.ndarray): 2D matrix of word weights
        """
        assert self.model is not None, 'Cannot assign weights, apply build() before trying to ' \
                                       'loading embedding weights '
        emb_layer = self.model.get_layer(name='word_embedding')
        assert emb_layer.output_dim == weights.shape[1], 'embedding vectors shape mismatch'
        emb_layer.set_weights([weights])


class MultiTaskIntentModel(IntentExtractionModel):
    """
    Multi-Task Intent and Slot tagging model (using tf.keras)

    Args:
        use_cudnn (bool, optional): use GPU based model (CUDNNA cells)
    """

    def __init__(self, use_cudnn=False):
        super().__init__()
        self.model = None
        self.word_length = None
        self.num_labels = None
        self.num_intent_labels = None
        self.word_vocab_size = None
        self.char_vocab_size = None
        self.word_emb_dims = None
        self.char_emb_dims = None
        self.char_lstm_dims = None
        self.tagger_lstm_dims = None
        self.dropout = None
        self.use_cudnn = use_cudnn

    def build(self,
              word_length,
              num_labels,
              num_intent_labels,
              word_vocab_size,
              char_vocab_size,
              word_emb_dims=100,
              char_emb_dims=30,
              char_lstm_dims=30,
              tagger_lstm_dims=100,
              dropout=0.2):
        """
        Build a model

        Args:
            word_length (int): max word length (in characters)
            num_labels (int): number of slot labels
            num_intent_labels (int): number of intent classes
            word_vocab_size (int): word vocabulary size
            char_vocab_size (int): character vocabulary size
            word_emb_dims (int, optional): word embedding dimensions
            char_emb_dims (int, optional): character embedding dimensions
            char_lstm_dims (int, optional): character feature LSTM hidden size
            tagger_lstm_dims (int, optional): tagger LSTM hidden size
            dropout (float, optional): dropout rate
        """
        self.word_length = word_length
        self.num_labels = num_labels
        self.num_intent_labels = num_intent_labels
        self.word_vocab_size = word_vocab_size
        self.char_vocab_size = char_vocab_size
        self.word_emb_dims = word_emb_dims
        self.char_emb_dims = char_emb_dims
        self.char_lstm_dims = char_lstm_dims
        self.tagger_lstm_dims = tagger_lstm_dims
        self.dropout = dropout

        words_input = tf.keras.layers.Input(shape=(None,), name='words_input')
        embedding_layer = tf.keras.layers.Embedding(self.word_vocab_size,
                                                    self.word_emb_dims, name='word_embedding')
        word_embeddings = embedding_layer(words_input)
        word_embeddings = tf.keras.layers.Dropout(self.dropout)(word_embeddings)

        # create word character input and embeddings layer
        word_chars_input = tf.keras.layers.Input(shape=(None, self.word_length),
                                                 name='word_chars_input')
        char_embedding_layer = tf.keras.layers.Embedding(self.char_vocab_size, self.char_emb_dims,
                                                         input_length=self.word_length,
                                                         name='char_embedding')
        # apply embedding to each word
        char_embeddings = char_embedding_layer(word_chars_input)
        # feed dense char vectors into BiLSTM
        char_embeddings = tf.keras.layers.TimeDistributed(
            tf.keras.layers.Bidirectional(self._rnn_cell(self.char_lstm_dims)))(char_embeddings)
        char_embeddings = tf.keras.layers.Dropout(self.dropout)(char_embeddings)

        # first BiLSTM layer (used for intent classification)
        first_bilstm_layer = tf.keras.layers.Bidirectional(
            self._rnn_cell(self.tagger_lstm_dims, return_sequences=True, return_state=True))
        first_lstm_out = first_bilstm_layer(word_embeddings)

        lstm_y_sequence = first_lstm_out[:1][0]  # save y states of the LSTM layer
        states = first_lstm_out[1:]
        hf, _, hb, _ = states  # extract last hidden states
        h_state = tf.keras.layers.concatenate([hf, hb], axis=-1)
        intents = tf.keras.layers.Dense(self.num_intent_labels, activation='softmax',
                                        name='intent_classifier_output')(h_state)

        # create the 2nd feature vectors
        combined_features = tf.keras.layers.concatenate([lstm_y_sequence, char_embeddings],
                                                        axis=-1)

        # 2nd BiLSTM layer for label classification
        second_bilstm_layer = tf.keras.layers.Bidirectional(self._rnn_cell(self.tagger_lstm_dims,
                                                                           return_sequences=True))(
            combined_features)
        second_bilstm_layer = tf.keras.layers.Dropout(self.dropout)(second_bilstm_layer)
        bilstm_out = tf.keras.layers.Dense(self.num_labels)(second_bilstm_layer)

        # feed BiLSTM vectors into CRF
        with tf.device('/cpu:0'):
            crf = CRF(self.num_labels, name='intent_slot_crf')
            labels = crf(bilstm_out)

        # compile the model
        model = tf.keras.Model(inputs=[words_input, word_chars_input],
                               outputs=[intents, labels])

        # define losses and metrics
        loss_f = {'intent_classifier_output': 'categorical_crossentropy',
                  'intent_slot_crf': crf.loss}
        metrics = {'intent_classifier_output': 'categorical_accuracy',
                   'intent_slot_crf': crf.viterbi_accuracy}

        model.compile(loss=loss_f,
                      optimizer=tf.train.AdamOptimizer(),
                      metrics=metrics)
        self.model = model

    def _rnn_cell(self, units, **kwargs):
        if self.use_cudnn:
            rnn_cell = tf.keras.layers.CuDNNLSTM(units, **kwargs)
        else:
            rnn_cell = tf.keras.layers.LSTM(units, **kwargs)
        return rnn_cell

    # pylint: disable=arguments-differ
    def save(self, path):
        """
        Save model to path

        Args:
            path (str): path to save model
        """
        super().save(path, ['use_cudnn'])


class Seq2SeqIntentModel(IntentExtractionModel):
    """
    Encoder Decoder Deep LSTM Tagger Model (using tf.keras)
    """

    def __init__(self):
        super().__init__()
        self.model = None
        self.vocab_size = None
        self.tag_labels = None
        self.token_emb_size = None
        self.encoder_depth = None
        self.decoder_depth = None
        self.lstm_hidden_size = None
        self.encoder_dropout = None
        self.decoder_dropout = None

    def build(self,
              vocab_size,
              tag_labels,
              token_emb_size=100,
              encoder_depth=1,
              decoder_depth=1,
              lstm_hidden_size=100,
              encoder_dropout=0.5,
              decoder_dropout=0.5):
        """
        Build the model

        Args:
            vocab_size (int): vocabulary size
            tag_labels (int): number of tag labels
            token_emb_size (int, optional): token embedding vector size
            encoder_depth (int, optional): number of encoder LSTM layers
            decoder_depth (int, optional): number of decoder LSTM layers
            lstm_hidden_size (int, optional): LSTM layers hidden size
            encoder_dropout (float, optional): encoder dropout
            decoder_dropout (float, optional): decoder dropout
        """
        self.vocab_size = vocab_size
        self.tag_labels = tag_labels
        self.token_emb_size = token_emb_size
        self.encoder_depth = encoder_depth
        self.decoder_depth = decoder_depth
        self.lstm_hidden_size = lstm_hidden_size
        self.encoder_dropout = encoder_dropout
        self.decoder_dropout = decoder_dropout

        words_input = tf.keras.layers.Input(shape=(None,), name='words_input')
        emb_layer = tf.keras.layers.Embedding(self.vocab_size, self.token_emb_size,
                                              name='word_embedding')
        benc_in = emb_layer(words_input)

        assert self.encoder_depth > 0, 'Encoder depth must be > 0'
        for i in range(self.encoder_depth):
            bencoder = tf.keras.layers.LSTM(self.lstm_hidden_size, return_sequences=True,
                                            return_state=True,
                                            go_backwards=True, dropout=self.encoder_dropout,
                                            name='encoder_blstm_{}'.format(i))(benc_in)
            benc_in = bencoder[0]
        b_states = bencoder[1:]
        benc_h, bene_c = b_states

        decoder_inputs = benc_in
        assert self.decoder_depth > 0, 'Decoder depth must be > 0'
        for i in range(self.decoder_depth):
            decoder = \
                tf.keras.layers.LSTM(
                    self.lstm_hidden_size,
                    return_sequences=True,
                    name='decoder_lstm_{}'.format(i))(decoder_inputs, initial_state=[benc_h,
                                                                                     bene_c])
            decoder_inputs = decoder
        decoder_outputs = tf.keras.layers.Dropout(self.decoder_dropout)(decoder)
        decoder_predictions = tf.keras.layers.TimeDistributed(
            tf.keras.layers.Dense(self.tag_labels, activation='softmax'),
            name='decoder_classifier')(decoder_outputs)

        self.model = tf.keras.Model(words_input, decoder_predictions)
        self.model.compile(optimizer=tf.train.AdamOptimizer(),
                           loss='categorical_crossentropy',
                           metrics=['categorical_accuracy'])



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import os
from six.moves import urllib
import numpy as np
from nlp_architect.utils.generic import license_prompt


class FastTextEmb:
    """
    Downloads FastText Embeddings for a given language to the given path.
    Arguments:
        path(str): Local path to copy embeddings
        language(str): Embeddings language
        vocab_size(int): Size of vocabulary
    Returns:
        Returns a dictionary and reverse dictionary
        Returns a numpy array with embeddings in emb_sizexvocab_size shape
    """

    def __init__(self, path, language, vocab_size, emb_dim=300):
        self.path = path
        self.language = language
        self.vocab_size = vocab_size
        self.emb_dim = emb_dim
        self.url = "https://s3-us-west-1.amazonaws.com/fasttext-vectors/wiki." + language + ".vec"

    def _maybe_download(self):
        """
        Download filename from url unless it's already in directory
        """
        # 1. Check if the file doesnt exist. Download and extract if it doesnt
        filename = "wiki." + self.language + ".vec"
        filepath = os.path.join(self.path, filename)
        link = "https://github.com/facebookresearch/fastText/blob/master/pretrained-vectors.md"
        if not os.path.exists(filepath):
            if license_prompt(filepath, link, self.path):
                print(
                    "Downloading FastText embeddings for " + self.language + " to " + filepath)
                urllib.request.urlretrieve(self.url, filepath)
                statinfo = os.stat(filepath)
                print(
                    "Sucessfully downloaded", filename, statinfo.st_size, "bytes")
            else:
                exit()
        else:
            print(
                "Found FastText embeddings for " + self.language + " at " + filepath)
        return filepath

    def read_embeddings(self, filepath):
        word2id = {}
        word_vec = []
        with open(filepath) as emb_file:
            for i, line in enumerate(emb_file):
                # Line zero has total words, emb dimensions
                if i == 0:
                    split_line = line.split()
                    assert len(split_line) == 2
                    assert self.emb_dim == int(split_line[1])
                # Rest of line are word, word_vec format
                else:
                    word, vector = line.rstrip().split(' ', 1)
                    vector = np.fromstring(vector, sep=' ')
                    # If norm is zero fill with 0.01
                    if np.linalg.norm(vector) == 0:
                        vector[0] = 0.01
                    assert vector.shape == (self.emb_dim, ), i
                    # Assign a token
                    word2id[word] = len(word2id)
                    word_vec.append(vector[None])
                # Check if your reached goal of vocab_size
                if i >= self.vocab_size:
                    break
        # Reverse dictionary
        id2word = {v: k for k, v in word2id.items()}
        # Dictionary just combines both id2word and word2id into one dict
        dico = Dictionary(id2word, word2id, self.language)
        # All word_vectors
        word_vec = np.concatenate(word_vec, 0)
        # Normalize the embeddings
        return dico, word_vec

    def load_embeddings(self):
        # Check if embeddings exist else download
        filepath = self._maybe_download()
        # Read embeddings
        dico, word_vec = self.read_embeddings(filepath)
        print("Completed loading embeddings for " + self.language)
        word_vec = np.float32(word_vec)
        return dico, word_vec


def get_eval_data(eval_path, src_lang, tgt_lang):
    """
    Downloads evaluation cross lingual dictionaries to the eval_path
    Arguments:
        eval_path: Path where cross-lingual dictionaries are downloaded
        src_lang : Source Language
        tgt_lang : Target Language
    Returns:
        Path to where cross lingual dictionaries are downloaded
    """
    eval_url = 'https://s3.amazonaws.com/arrival/dictionaries/'
    link = "https://github.com/facebookresearch/MUSE#ground-truth-bilingual-dictionaries"
    src_path = os.path.join(eval_path, '%s-%s.5000-6500.txt' % (src_lang, tgt_lang))
    filename = src_lang + '-' + tgt_lang + '.5000-6500.txt'
    if not os.path.exists(src_path):
        if license_prompt(src_path, link, src_path):
            os.system("mkdir -p " + eval_path)
            print("Downloading cross-lingual dictionaries for " + src_lang)
            urllib.request.urlretrieve(eval_url + filename, src_path)
            print("Completed downloading to " + eval_path)
        else:
            exit()
    return src_path


class Dictionary:
    """
    Merges word2idx and idx2word dictionaries
    Arguments:
        id2word dictionary
        word2id dictionary
        language of the dictionary
    Usage:
        dico.index(word) - returns an index
        dico[index] - returns the word
    """

    def __init__(self, id2word, word2id, lang):
        assert len(id2word) == len(word2id)
        self.id2word = id2word
        self.word2id = word2id
        self.lang = lang
        self.check_valid()

    def __len__(self):
        """
        Returns the number of words in the dictionary.
        """
        return len(self.id2word)

    def __getitem__(self, i):
        """
        Returns the word of the specified index.
        """
        return self.id2word[i]

    def __contains__(self, w):
        """
        Returns whether a word is in the dictionary.
        """
        return w in self.word2id

    def __eq__(self, y):
        """
        Compare the dictionary with another one.
        """
        self.check_valid()
        y.check_valid()
        if len(self.id2word) != len(y):
            return False
        return self.lang == y.lang and all(
            self.id2word[i] == y[i] for i in range(len(y)))

    def check_valid(self):
        """
        Check that the dictionary is valid.
        """
        assert len(self.id2word) == len(self.word2id)
        for i in range(len(self.id2word)):
            assert self.word2id[self.id2word[i]] == i

    def index(self, word):
        """
        Returns the index of the specified word.
        """
        return self.word2id[word]
